"""Extract fencing event registration stats from Fencing Time webpages."""

import asyncio
from concurrent.futures import ThreadPoolExecutor
from datetime import date, datetime, timedelta
from urllib.parse import urlparse, urljoin, urlencode
from lxml import html  # nosec ; Bandit suggests defusedxml but defusedxml.lxml is dead
from json.decoder import JSONDecodeError
import requests
from models import Event, EventPhase, Fencer, Tournament

# pylint: disable=too-few-public-methods ; I'm ok with that.


class Scraper:
    """Base class."""

    def __init__(self):
        """Initialize common args for scrapers."""
        self.tournament = None


class FTPScraper(Scraper):
    """Scraper for tournaments hosted on an FTP server.

    This reads the original Fencing Time results pages, hosted by individual
    tournament operators.
    """

    UPDATED_DATETIME_FORMAT = '%m/%d/%Y - %I:%M %p'
    EVENT_DATETIME_FORMAT = '%A, %B %d, %Y - %I:%M %p'

    def __init__(self, tournament_url):
        """Set up the scraper instance."""
        self.tournament_url = tournament_url
        super(FTPScraper, self).__init__()

    def scrape_tournament(self):
        """Get all tournament information."""
        try:
            results = requests.get(self.tournament_url)
        except requests.exceptions.MissingSchema:
            results = requests.get("http://{}".format(self.tournament_url))
        tournament_etree = html.fromstring(results.content)
        try:
            tournament_name = tournament_etree.xpath(
                '//span[@class="tournName"]/text()')[0]
            updated_str = (tournament_etree.xpath(
                '//span[@class="lastUpdate"]/text()')[0]
                           .replace('Last Updated:', '').strip())
            updated = datetime.strptime(updated_str, self.UPDATED_DATETIME_FORMAT)
        except IndexError:
            raise ScrapeError("Tournament info not found.")

        self.tournament = Tournament(name=tournament_name, url=results.url,
                                     updated=updated)

        # Get tournament events
        try:
            event_urls = tournament_etree.xpath(
                '//div[@id="schedule"]/table/tr/td/a[text()="View"]/@href')
        except IndexError:
            raise ScrapeError("No event schedule found.")

        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._scrape_events(event_urls))

        return self.tournament

    async def _scrape_events(self, event_urls):
        """Get event information asynchronously."""
        with ThreadPoolExecutor(max_workers=20) as executor:
            loop = asyncio.get_event_loop()
            futures = []

            for event_url in event_urls:
                if not urlparse(event_url).netloc:
                    event_url = urljoin(self.tournament.url, event_url)
                futures.append(loop.run_in_executor(executor, requests.get, event_url))

            for response in await asyncio.gather(*futures):
                event = self._parse_event(response)
                self.tournament.events.append(event)

            self.tournament.count_fencers()

    def _parse_event(self, event):
        """Extract useful strings from the event info."""
        event_tree = html.fromstring(event.content)
        event_details = event_tree.xpath('//span[@class="tournDetails"]/text()')
        try:
            event_name = event_details[0]
            event_time_str = event_details[1]
            event_time = datetime.strptime(event_time_str, self.EVENT_DATETIME_FORMAT)
        except IndexError:
            raise ScrapeError("Failed to interpret live results for event \"{}\"."
                              .format(event_details))

        # Get the event status
        if event_tree.xpath('//a[text()="Final Results"]'):
            fencers = event_tree.xpath(
                '//div[@id="finalResults"]/table/tr/td[2]/text()')
            fencers = [Fencer(f, True) for f in fencers]
            event_phase = EventPhase.FINISHED
        elif event_tree.xpath('//a[text()="Seeding"]'):
            fencers = event_tree.xpath(
                '//div[@id="Round1Seeding"]/table/tr/td[2]/text()')
            fencers = [Fencer(f, True) for f in fencers]
            event_phase = EventPhase.STARTED
        elif event_tree.xpath('//a[text()="Check-In Status"]'):
            event_phase = EventPhase.REGISTRATION
            fencers_checked_in = [
                bool(list(f))
                for f in event_tree.xpath('//div[@id="checkIn"]/table/tr/td[1]')]
            fencers = event_tree.xpath('//div[@id="checkIn"]/table/tr/td[2]/text()')
            fencers = [Fencer(f, ci) for (f, ci) in zip(fencers, fencers_checked_in)]

        return Event(name=event_name, time=event_time, phase=event_phase,
                     url=event.url, fencers=fencers, tournament=self.tournament)


class FTLiveScraper(Scraper):
    """Scraper for tournaments hosted on fencingtimelive.com.

    This reads the newer-style pages, centrally hosted by Fencing Time.
    """

    BASE_URL = 'https://fencingtimelive.com'
    TOURNAMENTS_URL = urljoin(BASE_URL, 'tournaments/search/data/advanced?{query}')
    TOURNAMENT_URL = urljoin(BASE_URL, 'tournaments/eventSchedule/{tournament_id}')
    FENCERS_URL = urljoin(BASE_URL, 'events/competitors/data/{event_id}')
    EVENT_URL = urljoin(BASE_URL, 'events/view/{event_id}')
    START_FORMAT = '%Y-%m-%dT%H:%M:%S.000Z'
    EVENT_DATETIME_FORMAT = '%A %B %d, %Y %I:%M %p'
    MAX_AGO = timedelta(days=7)
    MAX_AHEAD = timedelta(days=7)

    def list_tournaments(self, search=None, from_date=None, to_date=None):
        """Get a list of tournaments in FTLive."""
        if not from_date:
            from_date = date.today() - self.MAX_AGO
        if not to_date:
            to_date = date.today() + self.MAX_AHEAD
        args = {
            'tname': search or '',
            'from': from_date or '',
            'to': to_date or ''
        }
        url = self.TOURNAMENTS_URL.format(query=urlencode(args))
        try:
           tournaments = requests.get(url).json()
        except JSONDecodeError:
            raise ScrapeError("Failed to decode tournament list from URL {url}"
                              .format(url=url))
        return [{'start': datetime.strptime(t['start'], self.START_FORMAT),
                 'id': t['id'],
                 'name': t['name'],
                 'location': t['location']}
                for t in tournaments]

    def scrape_tournament(self, tournament_id):
        """Get all tournament information."""
        tournament_url = self.TOURNAMENT_URL.format(tournament_id=tournament_id)
        tournament_html = requests.get(tournament_url).content
        tournament_etree = html.fromstring(tournament_html)
        try:
            tournament_name = tournament_etree.xpath(
                '//div[@class="desktop tournName"]/text()')[0]
        except IndexError:
            raise ScrapeError("Tournament info not found.")
        self.tournament = Tournament(name=tournament_name, url=tournament_url,
                                     ftl_id=tournament_id)

        event_data = tournament_etree.xpath(
            "//tr[re:test(@id, 'ev_.*')]",
            namespaces={"re": "http://exslt.org/regular-expressions"})

        for event in event_data:
            self.tournament.events.append(self._parse_event(event))

        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._get_fencers())

        return self.tournament

    async def _get_fencers(self):
        """Get event information asynchronously."""
        def get_fencers(event):
            fencers_url = self.FENCERS_URL.format(event_id=event.ftl_id)
            response = requests.get(fencers_url).json()
            event.fencers = [
                Fencer(name=f['name'], ftl_id=f['id'],
                       is_checked_in=(f['status'] == 'CheckedIn'))
                for f in response if f['status'] != 'Scratched']
            return event

        with ThreadPoolExecutor(max_workers=20) as executor:
            loop = asyncio.get_event_loop()
            futures = []

            for event in self.tournament.events:
                futures.append(loop.run_in_executor(executor, get_fencers, event))

            await asyncio.gather(*futures)

            self.tournament.count_fencers()

    def _parse_event(self, event_etree):
        """Extract useful strings from the event info."""
        try:
            name = event_etree.getchildren()[1].text_content().strip()
            ftl_id = event_etree.attrib['data-href'].split('/')[-1]
            url = self.EVENT_URL.format(event_id=ftl_id)
            time_str = event_etree.getchildren()[0].text_content().strip()
            table = next(event_etree.iterancestors('table'))
            date_str = next(table.itersiblings('h5', preceding=True)).text_content()
            dt_str = '{} {}'.format(date_str, time_str)
            time = datetime.strptime(dt_str, self.EVENT_DATETIME_FORMAT)
            return Event(name=name, url=url, time=time, tournament=self.tournament,
                         phase=self._get_event_phase(event_etree), ftl_id=ftl_id)
        except (IndexError, ScrapeError) as exc:
            raise ScrapeError("Failed to interpret live results for event \"{}\". {}"
                              .format(name, exc))

    def _get_event_phase(self, event_etree):  # pylint: disable=no-self-use
        """Determine the state of the event."""
        if event_etree.xpath('td[3]/text()[contains(., "Finished")]'):
            event_phase = EventPhase.FINISHED
        elif event_etree.xpath('td[3]/text()[contains(., "Fencing")]'):
            event_phase = EventPhase.STARTED
        else:
            event_phase = EventPhase.REGISTRATION
        return event_phase


class ScrapeError(Exception):
    """A problem with scraping."""
